/*
    Copyright (c) 2002 Florian Wesch <fw@dividuum.de>. All Rights Reserved.     
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
*/    


#include <stdio.h>
#include "common.h"
#include "bfasm.h"

/* 
   TODO

   * Flexibleres Speichermanagement fr temporre Variablen
*/   

// Funktionen

BFAssemblerContext *BF_AsmInit(BFADDR freefortemp) {
    BFAssemblerContext *ctx;
    ctx = (BFAssemblerContext*)malloc(sizeof(BFAssemblerContext));

    if (ctx == NULL)
        fatalerror("Huch. keinen Speicher fr den Assembler?");

    ctx->looplevel = 0;
    ctx->first = NULL;
    ctx->cur = NULL;
    ctx->freefortemp = freefortemp;
    ctx->metainstructions = 0;

    return ctx;
}

void BF_AllocNewInst(BFAssemblerContext *ctx) {
    BFInstruction* new = (BFInstruction*)malloc(sizeof(BFInstruction));
    if (new == NULL)
        fatalerror("Huch. keinen Speicher fr neue Anweisung?");

    if (ctx->first == NULL) {
        ctx->first = new;
        ctx->cur = ctx->first;
        new->prev = NULL;
    } else {
        ctx->cur->next = new;
        new->prev = ctx->cur;
        ctx->cur = ctx->cur->next;
    }
    ctx->cur->next = NULL;
    ctx->metainstructions++;
}

void BF_RemoveInst(BFAssemblerContext *ctx, BFInstruction *inst) {
    if (inst->prev == NULL) {
        ctx->first = inst->next;
        inst->next->prev = NULL;        
    } else {
        inst->prev->next = inst->next;
        if (inst->next)
            inst->next->prev = inst->prev;
    }
    free(inst);
    ctx->metainstructions--;
}

void BF_AssemblerDump(BFAssemblerContext *ctx) {
    BFInstruction *inst = ctx->first;
    while (inst != NULL) {
        switch (inst->type) {
            case BFInst:
                printf("%c ", inst->val.bfinst);
                break;
            case BFMove:
                printf("TO(%d) ", inst->val.moveto);
                break;
            case BFSetTo:
                printf("SET(%d) ", inst->val.setto);
                break;
            case BFLoopS:
                printf("LOOPS(%p <-> %p) ", inst, inst->val.looppeer);
                break;
            case BFLoopE:
                printf("LOOPE(%p <-> %p) ", inst->val.looppeer, inst);
                break;
            case BFMoveRel:
                printf("RELMOVE(%d) ", inst->val.moverelative);
                break;
            case BFRecalibrateMP:
                printf("RECALIBRATE(%d) ", inst->val.newmp);
                break;
            case BFDebug:
                printf("DEBUG");
                break;
        }
        inst = inst->next;
    }
}

#define MoveFromTo(source, dest) {      \
    int i;                              \
    if (dest < source) {                \
        for (i=source; i>dest; i--) {   \
            fprintf(out, "<");          \
            ctx->instructions++;        \
        }                               \
    } else if (dest > source) {         \
        for (i=source; i<dest; i++) {   \
            fprintf(out, ">");          \
            ctx->instructions++;        \
        }                               \
    }                                   \
}

void BF_Assemble(BFAssemblerContext *ctx, FILE *out) {
    BFADDR lastmp=0; // MP steht am Anfang auf 0
    BFVALUE val;
    BFInstruction *inst = ctx->first;

    fprintf(out, "[ generated by bfcc "BFCCVERSION" ]\n");
    while (inst != NULL) {
        switch (inst->type) {
            case BFInst:
                fprintf(out, "%c", inst->val.bfinst);
                ctx->instructions++;
                break;
            case BFMove:
                MoveFromTo(lastmp, inst->val.moveto);
                lastmp = inst->val.moveto;
                break;
            case BFSetTo:
                fprintf(out, "[-]");
                ctx->instructions+=3;
                for (val = 0; val < inst->val.setto; val++) {
                    fprintf(out, "+");
                    ctx->instructions++;
                }
                break;
            case BFLoopS:
                fprintf(out, "[");
                ctx->instructions++;
                break;
            case BFLoopE:
                fprintf(out, "]");
                ctx->instructions++;
                break;
            case BFMoveRel:
                if (inst->val.moverelative > 0) {
                    int i;
                    for (i = 0; i < inst->val.moverelative; i++) 
                        fprintf(out, ">");
                } else if (inst->val.moverelative < 0) {
                    int i;
                    for (i = 0; i > inst->val.moverelative; i--) 
                        fprintf(out, "<");
                }
                lastmp += inst->val.moverelative;
                ctx->instructions += abs(inst->val.moverelative);
                break;
            case BFRecalibrateMP:
                lastmp = inst->val.newmp;
                break;
            case BFDebug:
                fprintf(out, "#");
                ctx->instructions++;
                break;
        }
        inst = inst->next;
    }
    MoveFromTo(lastmp, 0); // Am Ende wieder auf 0 zurck
}

BFADDR BF_Newvar(BFAssemblerContext *ctx) {
    BFADDR nextfree = ctx->freefortemp;
    ctx->freefortemp++;
    
    if (ctx->freefortemp >= MAXMEMSIZE)
        fatalerror("freefortemp >= %d\n", MAXMEMSIZE);
    
    //info("Neue Variable @%d\n", nextfree);

    return nextfree;
}

void BF_Removevar(BFAssemblerContext *ctx, BFADDR mem_to_free) {
    if (mem_to_free == ctx->freefortemp-1)
        ctx->freefortemp--;
    else
        fatalerror("Whoops: ctx->freefortemp-1 == %d, mem_to_free == %d", 
                    ctx->freefortemp-1,
                    mem_to_free);
    
    //info("Tte Variable @%d\n", mem_to_free);
}

void BF_StartLoop(BFAssemblerContext *ctx) {
    ctx->loopstack[ctx->looplevel] = ctx->cur;
    
    ctx->looplevel++;
    
    if (ctx->looplevel >= MAXLOOPLEVEL)
        fatalerror("Maxlooplevel berschritten");
    
}

void BF_EndLoop(BFAssemblerContext *ctx) {

    ctx->looplevel--;

    if (ctx->looplevel < 0)
        fatalerror("Looplevel < 0: Interer Fehler!?!");

    ctx->cur->val.looppeer = ctx->loopstack[ctx->looplevel];
    ctx->loopstack[ctx->looplevel]->val.looppeer = ctx->cur;
}

/*    
    to(dest)
        < oder >
*/    
BFINST1(to, BFADDR dest) {    
    BFMOVE(dest);
}

/*
    zero(dest)
        to(dest) {speicher auf 0 setzen} 
*/        
BFINST1(zero, BFADDR dest) { 
    BFCALL1(to, dest);
    BFSETTO(0);
}

/*
    left({constvalue})
        < * {constvalue}
*/
BFINST1(left, int constvalue) {
    BFMOVEREL(-constvalue);
}

/*
    right({constvalue})
        > * {constvalue}
*/
BFINST1(right, int constvalue) {
    BFMOVEREL(+constvalue);
}

/*
    print(mem)
        to(mem) . 
*/        
BFINST1(print, BFADDR mem) { 
    BFCALL1(to, mem);
    BFINST(BF_OPOUT); 
}

/*
    input(dest)
        to(dest) ,
*/        
BFINST1(input, BFADDR dest) { 
    BFCALL1(to, dest);
    BFINST(BF_OPIN); 
}


/*
    inc(dest)
        to(dest) + 
*/    
BFINST1(inc, BFADDR dest) {
    BFCALL1(to, dest);
    BFINST(BF_OPINCV);
}

/*
    dec(dest)
        to(dest) -
*/
BFINST1(dec, BFADDR dest) {
    BFCALL1(to, dest);
    BFINST(BF_OPDECV);
}

/*
    set_const(dest, {constvalue})
        to(dest) {speicher auf constvalue setzen} 
*/
BFINST2(set_const, BFADDR dest, BFVALUE constvalue) {
    BFCALL1(to, dest);
    BFSETTO(constvalue);
}




/*
    add_destroy(dest, value)
        to(value) [ inc(dest) dec(value) ]
*/
BFINST2(add_destroy, BFADDR dest, BFADDR value) {
    while_begin(value)
        BFCALL1(inc, dest);
        BFCALL1(dec, value);
    while_end(value);
}

/*
    substract_destroy(dest, value)
        to(value) [ dec(dest) dec(value) ]
*/
BFINST2(substract_destroy, BFADDR dest, BFADDR value) {
    while_begin(value)
        BFCALL1(dec, dest);
        BFCALL1(dec, value);
    while_end(value);
}


/*
    move_destroy(source, dest)
        zero(dest) add_destroy(dest, source)
*/
BFINST2(move_destroy, BFADDR source, BFADDR dest) {
    BFCALL1(zero, dest);
    BFCALL2(add_destroy, dest, source);
}

/*
    move2-destroy(source, dest1, dest2)
        zero(dest1) zero(dest2) to(source) [ inc(dest1) inc(dest2) dec(source) ]
*/        
BFINST3(move2_destroy, BFADDR source, BFADDR dest1, BFADDR dest2) {
    BFCALL1(zero, dest1);
    BFCALL1(zero, dest2);
    while_begin(source);
        BFCALL1(inc, dest1);
        BFCALL1(inc, dest2);
        BFCALL1(dec, source);
    while_end(source);
}

/*
    copy(source, dest)
        newvar(temp) move2-destroy(source, dest, temp) move-destroy(temp, source) removevar(temp)
*/
BFINST2(copy, BFADDR source, BFADDR dest) {
    BFADDR temp = BFNEWVAR;
        BFCALL3(move2_destroy, source, dest, temp);
        BFCALL2(move_destroy, temp, source);
    BFREMOVEVAR(temp);
}

/*
    mul(memory, value) 
        newvar(temp1) newvar(temp2) zero(temp1) to(memory) 
        [ copy(value, temp2) add-destroy(temp1, temp2) dec(memory) ] 
        move(temp1, memory) removevar(temp1) removevar(temp2)   
*/
BFINST2(mul, BFADDR memory, BFADDR value) {
    BFADDR temp1 = BFNEWVAR;
    BFADDR temp2 = BFNEWVAR;
        BFCALL1(zero, temp1);
        while_begin(memory);
            BFCALL2(copy, value, temp2);
            BFCALL2(add_destroy, temp1, temp2);
            BFCALL1(dec, memory);
        while_end(memory);
        BFCALL2(move_destroy, temp1, memory);
    BFREMOVEVAR(temp2);
    BFREMOVEVAR(temp1);
}

/*
    mulconst(memory, {constvalue})
        newvar(temp) setconst(temp, {constvalue}) mul(memory, temp) removevar(temp) 
*/        
BFINST2(mul_const, BFADDR memory, BFVALUE constvalue) {
    BFADDR temp = BFNEWVAR;
    BFCALL2(set_const, temp, constvalue);
    BFCALL2(mul, memory, temp);
    BFREMOVEVAR(temp);
}

/*
    not(mem)
        newvar(temp) setconst(temp, 1) if(mem, <zero(temp)>) move-destroy(temp, mem) removevar(temp)
*/
BFINST1(not, BFADDR memory) {
    BFADDR temp = BFNEWVAR;
    BFCALL2(set_const, temp, 1);
    if_begin(memory)
        BFCALL1(zero, temp);
    if_end;
    BFCALL2(move_destroy, temp, memory);
    BFREMOVEVAR(temp);
}

/*
    and(source1, source2, dest)
        zero(dest) if(source1, <copy(source2, dest)>)
*/
BFINST3(and, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFCALL1(zero, dest);
    if_begin(source1);
        BFCALL2(copy, source2, dest);
    if_end;
}

/*
    or(source1, source2, dest)
        newvar(temp) zero(dest) copy(source1, temp) add-destroy(dest, temp) 
        copy(source2, temp) add-destroy(dest, temp) removevar(temp)
*/        
BFINST3(or, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFADDR temp = BFNEWVAR;
    BFCALL1(zero, dest);
    BFCALL2(copy, source1, temp);
    BFCALL2(add_destroy, dest, temp);
    BFCALL2(copy, source2, temp);
    BFCALL2(add_destroy, dest, temp);
    BFREMOVEVAR(temp);
}

/*
    substractToNull-destroy(source1, source2)
        newvar(temp) and(source1, source2, temp)
        to(temp) [ dec(source1) dec(source2) and(source1, source2, temp) to(temp) ] removevar(temp)
*/
BFINST2(substractToNull_destroy, BFADDR source1, BFADDR source2) {
    BFADDR temp = BFNEWVAR;
    BFCALL3(and, source1, source2, temp);
    while_begin(temp);
        BFCALL1(dec, source1);
        BFCALL1(dec, source2);
        BFCALL3(and, source1, source2, temp);
    while_end(temp);
    BFREMOVEVAR(temp);
}

/*
    NotEqual-destroy(source1, source2, dest) 
        substractToNull-destroy(source1, source2)
        or(source1, source2, dest)
*/
BFINST3(notEqual_destroy, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFCALL2(substractToNull_destroy, source1, source2);
    BFCALL3(or, source1, source2, dest);
}

/*
    Equal-destroy(source1, source2, dest)
        NotEqual(source1, source2, dest) not(dest)
*/
BFINST3(Equal_destroy, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFCALL3(notEqual_destroy, source1, source2, dest);
    BFCALL1(not, dest);
}

/*
    Greater-destroy(source1, source2, dest)
        substractToNull-destroy(source1, source2)
        move-destroy(source1, dest)   
*/
BFINST3(Greater_destroy, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFCALL2(substractToNull_destroy, source1, source2);
    BFCALL2(move_destroy, source1, dest);
}

/*
    Less-destroy(source1, source2, dest)
        substractToNull-destroy(source1, source2)
        move-destroy(source2, dest)    
*/
BFINST3(Less_destroy, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFCALL2(substractToNull_destroy, source1, source2);
    BFCALL2(move_destroy, source2, dest);
}

/*
    GreaterOrEqual-destroy(source1, source2, dest) 
        inc(source1) Greater-destroy(source1, source2, dest)
*/        
BFINST3(GreaterOrEqual_destroy, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFCALL1(inc, source1);
    BFCALL3(Greater_destroy, source1, source2, dest);
}

/*
    LessOrEqual-destroy(source1, source2, dest) 
        inc(source2) Less-destroy(source1, source2, dest)
*/        
BFINST3(LessOrEqual_destroy, BFADDR source1, BFADDR source2, BFADDR dest) {
    BFCALL1(inc, source2);
    BFCALL3(Less_destroy, source1, source2, dest);
}

/*
    Thanks to Jeffry Johnston for his implementation of arrays in bf

    Lets give brainfuck a bufferoverflow feature ;-)

    Arrays in memory look like this:
    0VI#?#?#?#?#?#?#?#?     
    |||
    || \_ Index (0 = first Element)
    | \__ Value which will be copied to Arrayelement Index
     \___ All array must have the value 0 at address arraybase. 
          It will stop [<<] loops used to jump back to the beginning. 
*/     
   
BFINST1(ArrayInit, BFADDR arraybase) {
    BFCALL1(zero, arraybase);
}

BFINST3(ArraySet, BFADDR arraybase, BFADDR index, BFADDR source) {
    BFCALL2(copy, source, BFADDR_OFFSET(arraybase, 1)); // Prepare Value 
    BFCALL2(copy, index,  BFADDR_OFFSET(arraybase, 2)); // Prepare Index
    BFCALL1(to,   BFADDR_OFFSET(arraybase, 2));         // Set Memorypointer to Index
    BFLOOPS;                    // Loop until Index distributed among arrayelement metadata
        BFMOVEREL(+2);          // Set next Metadata to 0
        BFSETTO(0);
        BFMOVEREL(-2);
        BFINST(BF_OPDECV);      // Decrement index
        BFLOOPS;                // If decremented index > 0
            BFMOVEREL(+2);      // move decremented index from
            BFINST(BF_OPINCV);  //  current metadata to next metadata
            BFMOVEREL(-2);
            BFINST(BF_OPDECV);
        BFLOOPE;            
        BFINST(BF_OPINCV);      // Increment current metadata
        BFMOVEREL(+2);          // Repeat until next metadata = 0
    BFLOOPE; 

    // At this point, all metadata up to the metadata @Index is
    // filled with 1.
    //   _ always 0 (see above)             _ Memory Pointer
    //  /                                  / 
    // 0  1  2  3  4  5  6  7  8  9 10 11*12 13 14
    // 0  2  1  0  1  0  1  0  1  0  1  0  0  0  0
    //          0d 0m 1d 1m 2d 2m 3d 3m 4d 4m 5d 5m m=meta, d=data
    //                                         \_ Target
    
    BFMOVEREL(+1);              // Move to Target datacell 
    BFSETTO(0);                 // Set it to 0
    BFMOVEREL(-3);
    BFLOOPS;                    // Walk back
        BFMOVEREL(-2);
    BFLOOPE;
    BFMOVEREL(+1);              // Move to Value
    BFLOOPS;
        BFMOVEREL(+1);          // Move to Index (set to 1 or 0 now)
        BFLOOPS;                // Walk forward to end of 1 metadata
            BFMOVEREL(+2);  
        BFLOOPE;
        BFMOVEREL(+1);          // Move to final memorycell at array[targetIndex]
        BFINST(BF_OPINCV);      // Increment value
        BFMOVEREL(-3);
        BFLOOPS;                // Walk back
            BFMOVEREL(-2);
        BFLOOPE;
        BFMOVEREL(+1);          // Move to Value
        BFINST(BF_OPDECV);      // Decrement value
    BFLOOPE;

    // value moved from Value to array[targetIndex]

    BFMOVEREL(+1);              // Move to Index 
    BFINST(BF_OPDECV);
    BFMOVEREL(-1);

    BFRECALIBRATEMP(BFADDR_OFFSET(arraybase, 1));
}

BFINST3(ArrayGet, BFADDR arraybase, BFADDR index, BFADDR dest) {
    BFCALL2(copy, index,  BFADDR_OFFSET(arraybase, 2)); // Prepare Index
    BFCALL1(zero, BFADDR_OFFSET(arraybase, 1));         // Prepare Value
    BFCALL1(to,   BFADDR_OFFSET(arraybase, 2));         // Set Memorypointer to Index
    BFLOOPS;                    // Loop until Index distributed among arrayelement metadata
        BFMOVEREL(+2);          // Set next Metadata to 0
        BFSETTO(0);
        BFMOVEREL(-2);
        BFINST(BF_OPDECV);      // Decrement index
        BFLOOPS;                // If decremented index > 0
            BFMOVEREL(+2);      // move decremented index from
            BFINST(BF_OPINCV);  //  current metadata to next metadata
            BFMOVEREL(-2);
            BFINST(BF_OPDECV);
        BFLOOPE;            
        BFINST(BF_OPINCV);      // Increment current metadata
        BFMOVEREL(+2);          // Repeat until next metadata = 0
    BFLOOPE; 

    // At this point, all metadata up to the metadata @Index is
    // filled with 1.
    //   _ always 0 (see above)             _ Memory Pointer
    //  /                                  / 
    // 0  1  2  3  4  5  6  7  8  9 10 11*12 13 14
    // 0  2  1  0  1  0  1  0  1  0  1  0  0  0  0
    //          0d 0m 1d 1m 2d 2m 3d 3m 4d 4m 5d 5m m=meta, d=data
    //                                         \_ Target
    
    BFMOVEREL(+1);              // Move to Target datacell 

    // TESTME
    BFMOVEREL(+1);              // Zero Targets metadata
    BFLOOPS;
        BFINST(BF_OPDECV);
    BFLOOPE;
    BFMOVEREL(-1); 

    BFLOOPS;                    // move Target datacell to metadata of this and 
        BFMOVEREL(+1);          // previous metacell
        BFINST(BF_OPINCV);
        BFMOVEREL(-2);
        BFINST(BF_OPINCV);
        BFMOVEREL(+1); 
        BFINST(BF_OPDECV);
    BFLOOPE;
    BFMOVEREL(-1); 
    BFLOOPS;                    // move previous metacell back to datacell
        BFMOVEREL(+1);
        BFINST(BF_OPINCV);
        BFMOVEREL(-1);
        BFINST(BF_OPDECV);
    BFLOOPE;
    BFMOVEREL(+1);
    // end of TESTME

    BFLOOPS;
        BFMOVEREL(-3);
        BFLOOPS;                // Walk back
            BFMOVEREL(-2);
        BFLOOPE;
        BFMOVEREL(+1);          // Move to Value
        BFINST(BF_OPINCV);      // Increment value
        BFMOVEREL(+1);          // Move to Index (set to 1 or 0 now)
        BFLOOPS;                // Walk forward to end of 1 metadata
            BFMOVEREL(+2);  
        BFLOOPE;
        BFMOVEREL(+1);          // Move to final memorycell at array[targetIndex]
        BFINST(BF_OPDECV);      // Decrement value
    BFLOOPE;

    // TESTME
    BFMOVEREL(+1); 
    BFLOOPS;                    // Move metacell of target to datacell of target
        BFMOVEREL(-1);          // which is now 0. The value of target is
        BFINST(BF_OPINCV);      // therefor restored.
        BFMOVEREL(+1);
        BFINST(BF_OPDECV);
    BFLOOPE;
    BFMOVEREL(-1);
    // end of TESTME
            
    BFMOVEREL(-3);
    BFLOOPS;                    // Walk back
        BFMOVEREL(-2);
    BFLOOPE;
    BFMOVEREL(+1);              // Move to Value

    // value moved from array[targetIndex] to Value

    BFRECALIBRATEMP(BFADDR_OFFSET(arraybase, 1));

    BFCALL2(copy, BFADDR_OFFSET(arraybase, 1), dest); // Copy out value 
}

BFINST3(ArraySet_constindex, BFADDR arraybase, BFVALUE index, BFADDR source) {
    BFCALL2(copy, source, BFADDR_OFFSET(arraybase, 2 + index*2));
}

BFINST3(ArrayGet_constindex, BFADDR arraybase, BFVALUE index, BFADDR dest) {
    BFCALL2(copy, BFADDR_OFFSET(arraybase, 2 + index*2), dest);
}
